//  Giuseppe Vacanti (cosine science & computing bv)
//  January 10, 2003
// 
//   $Id: Exposure.cc,v 1.1.1.1 2004/05/20 13:50:06 eojero Exp $

#include<Observation.h>
#include<CalServer.h>

#include<sstream>
#include<iterator>
#include<iostream>
#include<algorithm>
#include<set>

namespace {
  Msg::Layer layer = Msg::App;
};


OdfIngest::Exposure::Exposure(const ::Exposure & oalexp,
			      const Instrument & instrument,
			      const Observation * observation)
  : start_(oalexp.actualStartTime()),
    end_(oalexp.actualEndTime()),
    instrument_(instrument),
//    ippvs_(oalexp.ippvs()),
    counter_(oalexp.counter()),
    scheduled_(oalexp.scheduled()),
    observation_(observation),
    camera_mode_(Xmm::MODE_NONE)
{}

OdfIngest::Exposure::Exposure(const ::Exposure * oalexp,
			      const Instrument & instrument,
			      const Observation * observation)
  : start_(oalexp->actualStartTime()),
    end_(oalexp->actualEndTime()),
    instrument_(instrument),
    //    ippvs_(oalexp->ippvs()),
    counter_(oalexp->counter()),
    scheduled_(oalexp->scheduled()),
    observation_(observation),
    camera_mode_(Xmm::MODE_NONE)
{}

OdfIngest::Exposure::Exposure(unsigned counter,
			      bool scheduled,
			      const STime& start,
			      const STime& end,
			      const Instrument & instrument,
			      const OdfDataSets & sets,
			      const Observation * observation)
  : start_(start),
    end_(end),
    instrument_(instrument),
    counter_(counter),
    scheduled_(scheduled),
    sets_(sets),
    observation_(observation),
    camera_mode_(Xmm::MODE_NONE)
{
  init_data_modes();
  init_calibration_modes();
  // FIXME: it isn't always meaningful to determine the camera mode.
  determine_camera_mode();
  modes_to_ippvs();
}


void OdfIngest::Exposure::init_data_modes() {

  // DATA MODE DETERMINATION
  // The aim here is to determine what data mode was used for each CCD
  // or window. This is in general a straightforward procedure, but
  // ther can be a few quirks that have to be taken into account, as
  // described below.
  
  // When one of the following data modes is encountered its name will
  // be retained only if there are no competing science modes. For
  // instance, the same exposure can contain both imaging and noise
  // data. When that is the case I want to remember that the data mode
  // is imaging.
  std::set<Xmm::DataMode> allowed_duplicates;
  allowed_duplicates.insert(Xmm::OFFSET_DATA);
  allowed_duplicates.insert(Xmm::NOISE_DATA);
  allowed_duplicates.insert(Xmm::DISCARDED_LINES_DATA);
  
  for(OdfDataSets::const_iterator k = sets_.begin();
      k != sets_.end(); ++k){
    const OdfDataSet & ds = *k;
    const Xmm::DataMode current_dmode = ds.dataMode();
  
    errstr << ds.basename() << " has data mode "
	   << current_dmode << Sas::message(layer, Msg::Verbose);

    // Only look at files that contain science data, not auxiliary
    // data sets.
    if(ds.type() != OdfDataSet::EVENT_LIST &&
       ds.type() != OdfDataSet::IMAGE)
      continue;
    
    // NOTE: The OAL has only the method ccdNr(), but for OM one
    // speaks of windows instead of CCDs.

    // I have no entry yet for this CCD/window. Store it and continue.
    if(data_modes_.find(ds.ccdNr()) == data_modes_.end()){
      data_modes_[ds.ccdNr()] = current_dmode;
      continue;
    }
        
    // I already have a data mode entry for this CCD/window.

    // One of the following can be true:
    //
    // [1] The current data mode is one of that I can skip if I have
    // already another entry.    
    if(allowed_duplicates.find(current_dmode) !=
       allowed_duplicates.end()) {
      errstr << "Looking at " << ds.basename()
	     << " I encountered another data mode entry for CCD/window "
	     << ds.ccdNr() << " and I'm ignoring it because it is "
	     << current_dmode
	     << Sas::message(layer, Msg::Noisy);
      continue;
    }

    // [2] The current data mode should not be skipped. Look at the
    // already existing entry, and if that one can be skipped the
    // current entry replaces it.    
    Xmm::DataMode existing_dmode = data_modes_[ds.ccdNr()];
    if(allowed_duplicates.find(existing_dmode) == allowed_duplicates.end())
      {
	errstr << "Looking at " << ds.basename()
	       << " I encountered a duplicate data mode entry for CCD/window "
	       << ds.ccdNr() << " that replaces the existing one."
	       << Sas::message(layer, Msg::Noisy);
	data_modes_[ds.ccdNr()] = current_dmode;
	continue;
      }
    
    // [3] Both the current and the existing data mode entries are to
    // be kept. Then all is ok if the two are identical, otherwise
    // there is a problem. We get here if the MOS central's CCD is
    // readout via the two nodes simultaneously. Then there will be
    // two datasets for the same CCD, and they must report the same
    // data mode.
    if(data_modes_[ds.ccdNr()] != current_dmode){
      errstr << "While looking at " << identifier()
	     << " I encounter what appears to be an inconsistent set of data modes for CCD/window "
	     << ds.ccdNr() << ". The dataset " << ds.basename()
	     << " reports to have datamode " << current_dmode
	     << " whereas I expected a value compatible with "
	     << data_modes_[ds.ccdNr()]
	     << ". This exposure will not have a valid data mode setting."
	     << Sas::warning("InconsistenDataModes");      
      data_modes_[ds.ccdNr()] = Xmm::DATA_MODE_NONE;
      continue;
    }
  }
}

void OdfIngest::Exposure::init_calibration_modes() {

  // CALIBRATION MODE DETERMINATION
  // The aim here is to determine for each CCD/window what calibration
  // mode would the CAL use. Not all data modes lead to a meaningful
  // calibration mode, hence a few datasets are not looked at.
  
  const DataModes & smodes = instrument_.science_eng_data_modes();
  for(OdfDataSets::const_iterator k = sets_.begin();
      k != sets_.end(); ++k){
    const OdfDataSet & ds = *k;

    // Only look at files that are science-related
    if(ds.type() != OdfDataSet::EVENT_LIST &&
       ds.type() != OdfDataSet::IMAGE)
      continue;
    
    // Only science modes (the test above could go)
    DataModes::const_iterator p = std::find(smodes.begin(), smodes.end(),
					    ds.dataMode());
    if(p != smodes.end()){
      Xmm::Mode current_mode = Xmm::MODE_NONE;
      // With the two true's we ask the CalServer not to issue any
      // warnings.
      calServer.setState(ds.name(), true, true);
      ChoiceStateVariable<Mode> * m = calServer.state()->mode();
      if(m->isUncertain()){
	errstr << "While looking at " << identifier()
	       << " I was not able to confidently determine the calibration mode for the dataset "
	       << ds.basename() << ". The answer I got is "
	       << m->get() << " but it is uncertain. This exposure will not have a valid instrument mode setting."
	       << Sas::warning("InconsistentCalibrationMode");
      } else {
	current_mode = m->get();
	errstr << ds.basename() << " has calibration mode "
	       << current_mode << Sas::message(layer, Msg::Sparse);
      }
      if(calibration_modes_.find(ds.ccdNr()) == calibration_modes_.end()){
	calibration_modes_[ds.ccdNr()] = current_mode;
	continue;
      }
      if(calibration_modes_[ds.ccdNr()] != current_mode){
	errstr << "While looking at " << identifier()
	       << " I encountered what appears to be an inconsistent set of calibration modes for CCD/window "
	       << ds.ccdNr() << ". The dataset " << ds.basename()
	       << " reports to have calibration mode " << current_mode
	       << " whereas I expected a value compatible with "
	       << calibration_modes_[ds.ccdNr()]
	       << ". This exposure will not have a valid instrument mode setting."
	       << Sas::warning("InconsistentCalibrationMode");
	calibration_modes_[ds.ccdNr()] = Xmm::MODE_NONE;
      }
    } else {
      errstr << ds.basename() << " is not a science dataset and it'll be ignored."
	     << Sas::message(Msg::App, Msg::Verbose);
    }
  }
}

void OdfIngest::Exposure::determine_camera_mode() {

  // CAMERA MODE DETERMINATION
  // The aim here is to determine the global camera mode
  // configuration, based on the available calibration modes. 
  
  // For MOS, the calibration mode of the CCD 1 (either node)
  // determines the instrument mode. One expects to see no more than
  // two different calibration/data modes, one for the central CCD,
  // and one for the external ones.
  
  // For RGS and PN the calibration mode for any of the CCDs gives the
  // instrument mode. Expect to see only one data mode.
  
  // For OM, the calibration mode for window 0 determines the
  // instrument mode. Except to see at most two data modes (IMI with
  // FAE).
  
  camera_mode_ = Xmm::MODE_NONE;
  bool can_determine_calibration_mode = true;

  std::set<Xmm::DataMode> data_modes_seen;
  std::map<unsigned, Xmm::DataMode>::const_iterator pdm;
  for(pdm = data_modes_.begin(); pdm != data_modes_.end(); ++pdm){
    if(pdm->second == Xmm::DATA_MODE_NONE){
      Sas::warning("FIXME", "At least one data mode is undefined.");
      can_determine_calibration_mode = false;
    } else
      data_modes_seen.insert(pdm->second);
  }

  std::set<Xmm::Mode> calibration_modes_seen;
  std::map<unsigned, Xmm::Mode>::const_iterator pm;
  for(pm = calibration_modes_.begin(); pm != calibration_modes_.end(); ++pm){
    if(pm->second == Xmm::MODE_NONE){
      Sas::warning("FIXME", "At least one calibration mode is undefined.");
      can_determine_calibration_mode = false;
    } else
      calibration_modes_seen.insert(pm->second);
  }

  if(calibration_modes_seen.size() == 0)
    can_determine_calibration_mode = false;

  if(can_determine_calibration_mode){    
    unsigned use_this_index = 1; // CCD 1, but window 0
    switch(instrument_.identifier()){
    case Xmm::OM:
      use_this_index = 0;
    case Xmm::EMOS1:
    case Xmm::EMOS2:
      if(calibration_modes_seen.size() > 2){
	Sas::warning("More than two calibration modes", "FIXME");
      } else {
	if(calibration_modes_.find(use_this_index) == calibration_modes_.end()){
	  Sas::warning("I needed mode for window/ccd x but found none.", "FIXME");
	} else {
	  camera_mode_ = calibration_modes_[use_this_index];
	}
      }
      break;
    case Xmm::RGS1:
    case Xmm::RGS2:
    case Xmm::EPN:
      if(calibration_modes_seen.size() > 1){
	Sas::warning("More than one calibration modes", "FIXME");
      } else {
	camera_mode_ = *(calibration_modes_seen.begin());
      }
      break;
      // FIXME:
    default: {}
    }
  } else
    Sas::warning("FIXME", "Cannot determine the camera mode.");
}


  

unsigned OdfIngest::Exposure::counter() const {
  return counter_;
}

bool OdfIngest::Exposure::scheduled() const {
  return scheduled_;
}

const STime & OdfIngest::Exposure::start_time() const {
  return start_;
}

const OdfIngest::Observation * OdfIngest::Exposure::observation() const {
  return observation_;
}

const STime & OdfIngest::Exposure::end_time() const {
  return end_;
}

std::string OdfIngest::Exposure::identifier(bool verbose) const {
  stringstream ss;
  if(verbose)
    ss << instrument_.nickname();
  ss << (scheduled_ ? "S" : "U") << std::setw(3) << std::setfill('0')
     << counter_;
  return ss.str();
}

const OdfIngest::Instrument & OdfIngest::Exposure::instrument() const {
  return instrument_;
}

void OdfIngest::Exposure::print_to(ostream & os) const {
  stringstream s;
  s << instrument_.nickname()
    << (scheduled_ ? "S" : "U")
    << setw(3) << setfill('0') << counter_;
  
  os << "// ---------------\n";
  os << "// Exposure " << s.str() << "\n";
  os << "// ---------------\n";
  os << "EXPOSURE\n";
  os << setw(3) << setfill('0') << counter_ << "\n";
  os << (scheduled_ ? "Y" : "N") << "\n";
  os << "SCIENCE\n";
  os << observation_->identifier() << s.str() << "\n";
  os << start_.toString(0) << "\n"; 
  os << start_.toString(0) << "\n"; 
  os << end_.toString(0) << "\n"; 
  os << end_.toString(0) << "\n";
  os << "CONFIGURATION " << instrument_.configuration() << "\n";
  Ippvs ip = ippvs();
  os << ip.size() << "\n";
  for(Ippvs::const_iterator K = ip.begin();
      K != ip.end(); ++K){
    if((*K).invalid())
      os << "// ";
    os << (*K).name() << " = " << (*K).value();
    if((*K).invalid())
      os << " // The extraction of this HK parameter failed";
    os << "\n";
  }  
}

void OdfIngest::Exposure::modes_to_ippvs() {
  Ippv camera_mode("MODE", Xmm::toString(camera_mode_));
  insert_ippv(camera_mode);
  for(std::map<unsigned, Xmm::DataMode>::const_iterator K = data_modes_.begin();
      K != data_modes_.end(); ++K){
    stringstream s;
    s << "DATA_MODE_" << (*K).first;
    Ippv ip(s.str(), Xmm::toString((*K).second));
    insert_ippv(ip);
  }
  for(std::map<unsigned, Xmm::Mode>::const_iterator K = calibration_modes_.begin();
	K != calibration_modes_.end(); ++K){
    stringstream s;
    s << "CALIBRATION_MODE_" << (*K).first;
    Ippv ip(s.str(), Xmm::toString((*K).second));
    insert_ippv(ip);
  }
}
  
bool OdfIngest::Exposure::can_extract_housekeeping_parameters() const {
  return (camera_mode_ != Xmm::DIAGNOSTIC &&
	  camera_mode_ != Xmm::MODE_NONE);
}
