//  Giuseppe Vacanti (cosine science & computing bv)
//  March  3, 2003
//
//   $Id: TimeRows.cc,v 1.2 2009/08/31 15:12:34 aibarra Exp $

#include<TimeRows.h>

#include<SmartDal>
#include<OdfingestUtilities.h>
#include<errstr.h>
/* AII gcc4 porting
 * error: ‘fabs’ is not a member of ‘std’
 */
#include <cmath>

#include<algorithm>
#include<exception>

namespace {
  class Compare {
  public:
    Compare() {}
    Compare(unsigned long value) : i_(value) {}
    bool operator()(const std::pair<unsigned long, STime> & p,
		    const STime & value) {
      return p.second < value;
    }
    bool operator()( const STime & value,
		     const std::pair<unsigned long, STime> & p
		     ) {
      return value < p.second;
    }
    bool operator()(const std::pair<unsigned long, STime> & p,
		    unsigned long value) {
      return p.first < value;
    }
    bool operator()(const std::pair<unsigned long, STime> & p) {
      return p.first == i_;
    }
  private:
    unsigned long i_;

  };

  class Compare2 {
  public:
    Compare2(const STime & time) : tref_(time) {}
    bool operator()(const std::pair<unsigned long, STime> & p) {
      return p.second >= tref_;
    }
  private:
    STime tref_;
  };

};

TimeRows::TimeRows(const SmartDal::TableH & tab) {
  using SmartDal::ColumnH;
  using SmartDal::ColumnDataH;

  ColumnH c = tab->column("TIME");
  ColumnDataH cd = c->data();
  for(unsigned k = 0; k < tab->rows(); ++k){
    std::string tstring = SmartDal::get_string_row(cd, k);
    try {
      SetSTimeException xx;
      push_back(std::make_pair(k,STime(tstring)));
    } catch (STime::InvalidTime e) {
      errstr << c->qualifiedName() << " at row " << k
	     << " contains an invalid time string: "
	     << tstring << Sas::warning("InvalidTimeString");
    }
  }
}
TimeRows::TimeRows(const SmartDal::TableH & tab,
		   const std::vector<DataFilter::RowRange>& rr) {
  using SmartDal::ColumnH;
  using SmartDal::ColumnDataH;

  ColumnH c = tab->column("TIME");
  ColumnDataH cd = c->data();
  for(unsigned k = 0; k < tab->rows(); ++k){
    bool skip = true;
    for(unsigned h = 0; h < rr.size(); ++h)
      if(k >= rr[h].startRow && k <= rr[h].stopRow){
	skip = false;
	break;
      }
    if(skip){
//       errstr << c->qualifiedName() << ": row " << k
// 	     << " has been marked as excluded."
// 	     << Sas::message(Msg::App, Msg::Verbose);
      continue;
    }
    std::string tstring = SmartDal::get_string_row(cd, k);
    try {
      SetSTimeException xx;
      push_back(std::make_pair(k,STime(tstring)));
    } catch (STime::InvalidTime e) {
      errstr << c->qualifiedName() << " at row " << k
	     << " contains an invalid time string: "
	     << tstring << Sas::warning("InvalidTimeString");
    }
  }
}


TimeRows::const_iterator TimeRows::lower_bound(const STime & time) const {
  Compare cmp;
  XX p = std::lower_bound(begin(), end(), time, cmp);
  return p;
}

TimeRows::const_iterator TimeRows::upper_bound(const STime & time) const {
  Compare cmp;
  XX p = std::upper_bound(begin(), end(), time, cmp);
  return p;
}

TimeRows::const_iterator TimeRows::begin() const {
  return std::vector<std::pair<unsigned long, STime> >::begin();
}

TimeRows::const_iterator TimeRows::end() const {
  return std::vector<std::pair<unsigned long, STime> >::end();
}

unsigned TimeRows::index(const_iterator i) const {
  if(i == end())
    errstr << "TimeRows: trying to get the index of a non-existing row."
	   << Sas::error("IteratorOutOfRange");
  return (*i).first;
}

STime TimeRows::totime(const_iterator i) const {
  if(i == end())
    errstr << "TimeRows: trying to get the time of a non-existing row."
	   << Sas::error("IteratorOutOfRange");
  return (*i).second;
}

// I'm looking for a time record earlier than time. This is the HK
// (non periodic) time record to be used to extract HK values. This is
// then the current (active) HK value.
TimeRows::const_iterator TimeRows::active_record(const STime & time) const {
  const_iterator ret = end();
  const_iterator low = lower_bound(time);

  // If lower_bound returns begin(), then there is no record earlier
  // than time.
  if(low == begin()){
    ret = end();
  } else {
    STime ntime = totime(low - 1);
    ret = lower_bound(totime(low - 1));
  }
  return ret;
}

std::pair<long, long> TimeRows::active_range(const STime & time) const {
  std::pair<long,long> ret = std::make_pair(-1,-1);

  const_iterator i = active_record(time);
  if(i != end()){
    STime trange = totime(i);

    bool have_first = false;
    bool have_second = false;
    const_iterator low, upp;
    for(const_iterator k = begin(); k != end(); ++k){
      double dt = std::fabs(trange - totime(k));
      if(dt < 1e-9 && !have_first){
	have_first = true;
	low = k;
      }
      if(have_first && dt > 1e-9){
	upp = k;
	have_second = true;
	break;
      }
    }

    ret.first = index(low);
    if(have_second)
      ret.second = index(upp);
    else
      ret.second = index(end() - 1) + 1;
  } else
    errstr << "No active range: earliest time later than start.\n"
	   << "Input time: " << time << "\n"
	   << "Earliest available time: " << totime(begin())
	   << Sas::error("NoActiveRange");

  return ret;
}
